//---------------------------------------------------------------------------
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#pragma hdrstop

#include "Assembler.h"
#include "ObjectFile.h"

//---------------------------------------------------------------------------
#pragma package(smart_init)

struct _Message
{
  long num;
  const char *text;
};

struct _Message _Errors[] = {
  {100, "Cannot open file: \"%s\""},
  {101, "Number of lines in intermediate file exceeds %s"},
  {102, "Number of files included exceeds %s"},
  {103, "Unrecognized preprocessor directive \"#%s\""},
  {104, "Filename expected"},
  {105, "File already included: \"%s\""},
  {106, "Identifier expected"},
  {107, "Too many defines (%s entries)"},
  {200, "Line is too long (max of %s chars)"},
  {201, "Label is too long (max of %s chars)"},
  {202, "Label required for ':'"},
  {203, "Invalid character '%s' in label"},
  {204, "Label cannot start with a number"},
  {205, "Duplicate label \"%s\" conflicts with previous definition"},
  {206, "Symbol table full (%s entries)\n"},
  {207, "Output buffer overflowed (max of %s bytes)"},
  {208, "Invalid character '%s' in external symbol"},
  {209, "External symbol used with non-arithmetic operator"},
  {210, "Multiple external symbols used in expression"},
  {211, "Cannot use external symbols in data declarations"},
  {212, "External symbol cannot be used as 8 bit constant"},
  {213, "Relative jump too long"},
  {214, "Relocatable code cannot be generated for local expression"},
  {215, "External symbol cannot be the target of a relative jump"},
  {216, "Symbol not defined: \"%s\""},
  {217, "Relative symbol cannot be used as 8 bit constant"},
  {218, "Cannot pad backwards"},
  {219, "Cannot open binary file: \"%s\""},
  {220, "Local label too long when expanded"},
  {221, "No top-level public label for local label"},
  {222, "Data declaration size must evaluate to a constant"},
  {223, "Pad value must evaluate to a constant"},
  {300, "Invalid expression: \"%s\""},
  {301, "Division by 0"},
  {302, "No opening '('"},
  {303, "No closing ')'"},
  {304, "Invalid character constant in expression"},
  {305, "Unrecognized directive: \".%s\""},
  {306, "Value or expression expected"},
  {307, "Invalid escape sequence: \\%s"},
  {308, "No closing quote in string"},
  {400, "Unable to open instruction table file \"%s\""},
  {401, "Error parsing instruction table file"},
  {402, "Opcode size must be between 1 and %s bytes"},
  {403, "Line exceeds %s characters"},
  {404, "Unrecognized instruction: \"%s\""},
  {405, "Unrecognized argument format for \"%s\" instruction"},
  {406, "Maximum of %s instructions in table exceeded"},
  {407, "External relocation not supported for instruction"},
};

struct _Message _Warnings[] = {
  {200, "Duplicate label \"%s\""},
  {201, "Symbol table entry truncated to 16 bits"},
  {300, "Value truncated to 16 bits"},
  {301, "Value truncated to 8 bits"},
  {400, "Argument value truncated to 16 bits"},
  {401, "Argument value truncated to 8 bits"},
  {402, "Illegal bit value"},
};

CAssembler assembler;

void CAssembler::Error(enum AsmError n, const char *s)
{
  static char last[256] = {'\0'};
  char message[256],
       line[256];
  int i;

  for(i = 0; _Errors[i].num != n; i++) {}

  sprintf(message, _Errors[i].text, s);
  sprintf(line, "Error #%i in %s, line %i: %s", n,
          Filename(), LineNumber(), message);
  if (strcmp(last, line))
  {
    printf("%s\n", line);
    _errors++;
  }
  strcpy(last, line);
}

void CAssembler::ErrorI(enum AsmError n, int i)
{
  char s[16];

  sprintf(s, "%i", i);
  Error(n, s);
}

void CAssembler::Warning(enum AsmWarning n, const char *s)
{
  static char last[256] = {'\0'};
  char message[256],
       line[256];
  int i;

  for(i = 0; _Warnings[i].num != n; i++) {}

  sprintf(message, _Warnings[i].text, s);
  sprintf(line, "Warning #%i in %s, line %i: %s", n,
          Filename(), LineNumber(), message);
  if (strcmp(last, line))
  {
    printf("%s\n", line);
    _warnings++;
  }
  strcpy(last, line);
}

void CAssembler::Message(const char *m, const char *s)
{
  char message[256];

  sprintf(message, m, s);
  printf("%s\n", message);
}

void CAssembler::NextLine()
{
  _linenumber++;
}

char *CAssembler::Filename()
{
  if (*_filename != '\0')
    return _filename;

  return _preprocessor.FileName(_linenumber);
}

int CAssembler::LineNumber()
{
  if (*_filename != '\0')
    return _linenumber;

  return _preprocessor.LineNum(_linenumber);
}

int CAssembler::PCounter()
{
  return _pcounter;
}

bool CAssembler::Errors()
{
  return _errors;
}

char *CAssembler::LastLocal()
{
  return _lastlocal;
}

bool CAssembler::Assemble(const char *tfile, const char *fname,
                          const char *oname, bool werror)
{
  const char TMPFILE[] = {"$asmtmp$.$$$"};
  char line[MAXLINE],
       *p, *t,
       sym[MAXSYM],
       buf[MAXSYM];
  FILE *fp;
  int i, n;
  bool exp;

  strcpy(_filename, tfile);
  _instruction.LoadTable(tfile);

  _errors = 0;
  _warnings = 0;
  *_filename = '\0';
  _linenumber = 0;
  Message("Assembling \"%s\"", fname);
  fp = fopen(TMPFILE, "w");
  if (fp == NULL)
  {
    Error(PPOPEN, TMPFILE);
    return false;
  }
  _preprocessor.Process(fname, fp, true);
  fclose(fp);

  if (_errors)
    return false;

  fp = fopen(TMPFILE, "r");
  if (fp == NULL)
  {
    Error(PPOPEN, TMPFILE);
    return false;
  }
  Message("First pass...", "");
  _pcounter = 0;
  *_lastlocal = '\0';
  while (!feof(fp))
  {
    _linenumber++;
    *line = '\0';
    fgets(line, sizeof(line), fp);
    if ((feof(fp)) && (strlen(line) == 0))
      break;
    if ((p = strchr(line, '\n')) != NULL)
      *p = '\0';
    else if ((strchr(line, '\n') == NULL) && (!feof(fp)))
      {
        ErrorI(PLINELEN, MAXLINE - 1);
        while (strchr(line, '\n') == NULL)
          fgets(line, sizeof(line), fp);
        continue;
      }
    if ((p = strchr(line, ';')) != NULL)
      *p = '\0';
    while ((p = strchr(line, '\t')) != NULL)
      *p = ' ';

    p = line;
    i = 0;
    exp = false;
    if ((*p != ' ') && (*p != '\0') && (*p != '.'))
    {
      while ((*p != ':') && (*p != '=') && (*p != ' ') && (*p != '\0'))
      {
        if (i == MAXSYM - 1)
        {
          ErrorI(PSYMLEN, MAXSYM - 1);
          break;
        }
        if ((i == 0) && (isdigit(*p)))
        {
          Error(PSYMNUM, "");
          break;
        }
        if ((!isalpha(*p)) && (!isdigit(*p)) &&
            (*p != '_') && (*p != '@'))
        {
          ErrorI(PSYMCHAR, p[0]);
          break;
        }
        sym[i++] = *p++;
      }
      if (i == 0)
      {
        Error(PSYMREQ, "");
        break;
      }
      sym[i] = '\0';
      if (*p == ':')
      {
        p++;
        if (*p == ':')
        {
          p++;
          exp = true;
        }
      }
      if (strchr(p, '=') == NULL)
      {
        _symtable.Add(sym, _pcounter, false, true, exp, NULL);
        if (*sym != '@')
          strcpy(_lastlocal, sym);
      }
      else
        if (*p != '\0')
        {
          while((*p != '\0') && (*p != '='))
            p++;
          if (*p == '\0')
          {
            _symtable.Add(sym, _pcounter, false, true, false, NULL);
            if (*sym != '@')
              strcpy(_lastlocal, sym);
          }
          else
          {
            n = Parse(++p, NULL, 0);
            _symtable.Add(sym, n, false, Local(p, false),
                          false, GetExtRef(p, buf));
            *p = '\0';
          }
        }
    }

    n = 0;
    while (*p != '\0')
    {
      if ((t = strchr(p, '\\')) != NULL)
        *t = '\0';
      while (*p == ' ')
        p++;
      if (*p == '.')
      {
        n = Directive(++p, true, false);
        if (n == -1)
          break;
        _pcounter += n;
      }
      else
      {
        _pcounter += _instruction.GetSize(p, buf);
        _pcounter += External(buf, false);
      }
      if (t != NULL)
        p = t + 1;
      else
          break;
    }
    if (n == -1)
      break;
  }
  fclose(fp);

  if (!_errors)
  {
    _linenumber = 0;
    fp = fopen(TMPFILE, "r");
    if (fp == NULL)
    {
      Error(PPOPEN, TMPFILE);
      return false;
    }

    Message("Second pass...", "");
    _pcounter = 0;
    *_lastlocal = '\0';
    objectfile.New();
    while (!feof(fp))
    {
      _linenumber++;
      *line = '\0';
      fgets(line, sizeof(line), fp);
      if ((feof(fp)) && (strlen(line) == 0))
        break;
      if ((p = strchr(line, '\n')) != NULL)
        *p = '\0';
      if ((p = strchr(line, ';')) != NULL)
        *p = '\0';
      while ((p = strchr(line, '\t')) != NULL)
        *p = ' ';

      p = line;
      i = 0;
      if ((*p != ' ') && (*p != '\0') && (*p != '.'))
      {
        while ((*p != ':') && (*p != '=') && (*p != ' ') && (*p != '\0'))
          p++;
        if (*p == ':')
        {
          strcpy(sym, line);
          sym[p - line] = '\0';
          if (*sym != '@')
            strcpy(_lastlocal, sym);
          p++;
          if (*p == ':')
            p++;
        }
        else
          if (strchr(p, '=') != NULL)
            continue;
      }

      n = 0;
      while (*p != '\0')
      {
        if ((t = strchr(p, '\\')) != NULL)
          *t = '\0';
        while (*p == ' ')
          p++;
        if (*p == '.')
        {
          n = Directive(++p, false, true);
          if (n == -1)
            break;
          _pcounter += n;
        }
        else
          _pcounter += _instruction.WriteOpcode(p);
        if (t != NULL)
          p = t + 1;
        else
            break;
      }
      if (n == -1)
        break;
    }
    fclose(fp);
  }
  remove(TMPFILE);

  sprintf(line, "Assembly complete - %i error(s), %i warning(s)",
          _errors, _warnings);
  Message(line, "");
  if (!(_errors || (_warnings && werror)))
    if (!objectfile.Write(oname, _symtable))
      return false;

  return !(_errors && (_warnings && werror));
}

int CAssembler::Parse(const char *c, int *e, int o)
{
  enum {OPS = 0, OPADD, OPSUB, OPMUL, OPDIV, OPEQ, OPNEQ, OPLT,
        OPGT, OPLTE, OPGTE, OPAND, OPOR, OPXOR, OPSL, OPSR} op;
  char expr[MAXLINE],
       *end;
  int  n, v,
       i, s;
  bool setop, ext;

  n = 0;
  i = 0;
  op = OPS;
  ext = false;
  while (c[i] != '\0')
  {
    setop = false;
    v = 0;
    strcpy(expr, c);
    while((c[i] != '\0') && (c[i] == ' '))
      i++;
    if ((isalpha(c[i])) || (c[i] == '_') || (c[i] == '@'))
    {
      s = i;
      while ((isalpha(c[i])) || (isdigit(c[i])) ||
             (c[i] == '_') || (c[i] == '@'))
        i++;
      expr[i] = '\0';
      v = _symtable.Get(&expr[s]) + (o * _symtable.Local(&expr[s]));
      if (_symtable.External(&expr[s]))
      {
        if (ext)
        {
          Error(PMULTIEXT, "");
          return 0;
        }
        ext = true;
      }
    }
    else if (isdigit(c[i]))
    {
      v = strtol(&expr[i], &end, 10);
      i = end - expr;
    }
    else if (c[i] == '$')
    {
      i++;
      if (!(isdigit(c[i])) && !(isalpha(c[i])))
        v = _pcounter + o;
      else
      {
        v = strtol(&expr[i], &end, 16);
        i = end - expr;
      }
    }
    else if (c[i] == '%')
    {
      v = strtol(&expr[++i], &end, 2);
      i = end - expr;
    }
    else if (c[i] == '\'')
    {
      if (c[++i] == '\0')
      {
        Error(PCHARCONST, "");
        break;
      }
      if (c[i] == '\\')
      {
        switch (c[++i])
        {
          case '\\':
            v = '\\';
            break;
          case '\'':
            v = '\'';
            break;
          case '"':
            v = '"';
            break;
          case '0':
            v = 0;
            break;
          default:
            expr[0] = c[i];
            expr[1] = '\0';
            Error(PINVESC, expr);
        }
        if (c[i] == '\0')
        {
          Error(PCHARCONST, "");
          break;
        }
      }
      else
        v = c[i];
      if (c[++i] != '\'')
        Error(PCHARCONST, "");
      else
        i++;
    }
    else if (c[i] == '(')
    {
      s = i;
      if (External(&expr[i + 1], false))
        ext = true;
      v = Parse(&expr[i + 1], &i, o);
      if (i == s)
        Error(PNOCLOSEP, "");
    }
    else if (c[i] == ')')
    {
      if (e == NULL)
        Error(PNOOPENP, "");
      else
        *e += (i + 2);
      return n;
    }
    else if ((c[i] == '=') && (c[i + 1] == '='))
    {
      op = OPEQ;
      i++;
    }
    else if ((c[i] == '!') && (c[i + 1] == '='))
    {
      op = OPNEQ;
      i += 2;
    }
    else if ((c[i] == '<') && (c[i + 1] == '='))
    {
      op = OPLTE;
      i += 2;
    }
    else if ((c[i] == '>') && (c[i + 1] == '='))
    {
      op = OPGTE;
      i += 2;
    }
    else if ((c[i] == '<') && (c[i + 1] == '<'))
    {
      op = OPSL;
      i += 2;
    }
    else if ((c[i] == '>') && (c[i + 1] == '>'))
    {
      op = OPSR;
      i += 2;
    }
    else
    {
      switch (c[i++])
      {
        case '+':
          op = OPADD; break;
        case '-':
          op = OPSUB; break;
        case '*':
          op = OPMUL; break;
        case '/':
          op = OPDIV; break;
        case '<':
          op = OPLT;  break;
        case '>':
          op = OPGT;  break;
        case '&':
          op = OPAND; break;
        case '|':
          op = OPOR;  break;
        case '^':
          op = OPXOR; break;
        default:
          Error(PILLEXPR, c);
      }
      setop = true;
    }
    if (setop == false)
    {
      switch (op)
      {
        case OPS:
          n = v;  break;
        case OPADD:
          n += v; break;
        case OPSUB:
          n -= v; break;
        case OPMUL:
          n *= v; break;
        case OPDIV:
          if (v == 0)
            Error(PDIVZERO, "");
          else
            n /= v;
          break;
        case OPEQ:
          n = (n == v); break;
        case OPNEQ:
          n = (n != v); break;
        case OPLT:
          n = (n < v);  break;
        case OPGT:
          n = (n > v);  break;
        case OPLTE:
          n = (n <= v); break;
        case OPGTE:
          n = (n >= v); break;
        case OPAND:
          n = (n & v);  break;
        case OPOR:
          n = (n | v);  break;
        case OPXOR:
          n = (n ^ v);  break;
        case OPSL:
          n = (n << v); break;
        case OPSR:
          n = (n >> v); break;
      }
      if ((ext == true) && (op != OPS) && (op != OPADD) && (op != OPSUB))
      {
        Error(PEXTINAR, "");
        return 0;
      }
    }
    while((c[i] != '\0') && ((c[i] == ' ') || (c[i] == '\t')))
      i++;
  }

  return n;
}

bool CAssembler::External(const char *c, bool obj)
{
  char expr[MAXLINE];
  int  i, s;

  i = 0;
  while (c[i] != '\0')
  {
    strcpy(expr, c);
    while((c[i] != '\0') && (!isalpha(c[i])) &&
          (c[i] != '_') && (c[i] != '@'))
      i++;
    if ((isalpha(c[i])) || (c[i] == '_') || (c[i] == '@'))
    {
      s = i;
      while ((isalpha(c[i])) || (isdigit(c[i])) ||
             (c[i] == '_') || (c[i] == '@'))
        i++;
      expr[i] = '\0';
      if (_symtable.External(&expr[s]))
      {
        if (obj)
          objectfile.AddExt(&expr[s]);
        return true;
      }
    }
  }

  return false;
}

bool CAssembler::Local(const char *c, bool obj)
{
  const int OFFSET = 9;
  int i, o;

  i = Parse(c, NULL, 0);
  o = Parse(c, NULL, OFFSET);
  if (o == i)
    return false;
  if (o - OFFSET != i)
    Error(PGENRELOC, "");
  if (obj)
    objectfile.AddLocal();

  return true;
}

bool CAssembler::ExtRef(const char *c, bool obj)
{
  char expr[MAXLINE];
  int  i, s;

  i = 0;
  while (c[i] != '\0')
  {
    strcpy(expr, c);
    while((c[i] != '\0') && (!isalpha(c[i])) &&
          (c[i] != '_') && (c[i] != '@'))
      i++;
    if ((isalpha(c[i])) || (c[i] == '_') || (c[i] == '@'))
    {
      s = i;
      while ((isalpha(c[i])) || (isdigit(c[i])) ||
             (c[i] == '_') || (c[i] == '@'))
        i++;
      expr[i] = '\0';
      if (_symtable.GetExt(&expr[s]) != NULL)
      {
        if (obj)
          objectfile.AddExt(_symtable.GetExt(&expr[s]));
        return true;
      }
    }
  }

  return false;
}

char *CAssembler::GetExtRef(const char *c, char *e)
{
  char expr[MAXLINE];
  int  i, s;

  i = 0;
  while (c[i] != '\0')
  {
    strcpy(expr, c);
    while((c[i] != '\0') && (!isalpha(c[i])) &&
          (c[i] != '_') && (c[i] != '@'))
      i++;
    if ((isalpha(c[i])) || (c[i] == '_') || (c[i] == '@'))
    {
      s = i;
      while ((isalpha(c[i])) || (isdigit(c[i])) ||
             (c[i] == '_') || (c[i] == '@'))
        i++;
      expr[i] = '\0';
      if (_symtable.External(&expr[s]))
      {
        strcpy(e, &expr[s]);
        return e;
      }
    }
  }

  return NULL;
}

int CAssembler::Directive(char *c, bool addsym, bool output)
{
  FILE *fp;
  const char *dirs[] = {"ext", "db", "byte", "dw", "word", "pad",
                "incbin", "end", "dbs", "bytes", "dws", "words"};
  char sym[MAXSYM],
       *e;
  int i, n, t;

  t = 0;
  for(i = 0; isalpha(c[i]); i++) {}
  for(n = i; c[n] == ' '; n++) {}
  if (c[n] == '\0')
  {
    c[i] = '\0';
    for(n = 0; strlen(dirs[n]) > 0; n++)
      if (!stricmp(c, dirs[n]))
        break;
    if (*dirs[n] == '\0')
      Error(PUNREQDIR, c);
    else
      if (stricmp(c, "end"))
        Error(PEXPREXP, "");
    return 0;
  }
  c[i++] = '\0';
  if (!stricmp(c, "ext"))
  {
    while (c[i] == ' ')
      i++;
    for(n = 0; ((c[i] != ' ') && (c[i] != '\0')); n++, i++)
    {
      if (n == MAXSYM - 1)
      {
        ErrorI(PSYMLEN, MAXSYM - 1);
        break;
      }
      if ((n == 0) && (isdigit(c[i])))
      {
        Error(PSYMNUM, "");
        break;
      }
      if ((!isalpha(c[i])) && (!isdigit(c[i])) && (c[i] != '_'))
      {
        ErrorI(PESYMCHAR, c[i]);
        break;
      }
      sym[n] = c[i];
    }
    sym[n] = '\0';
    if (addsym)
      _symtable.Add(sym, 0, true, false, false, NULL);
    return 0;
  }
  else if ((!stricmp(c, "db")) || (!stricmp(c, "byte")))
  {
    while (c[i] != '\0')
    {
      while (c[i] == ' ')
        i++;
      if (c[i] == '\0')
        break;
      if (c[i] == '\"')
      {
        if ((e = strchr(&c[++i], '\"')) == NULL)
        {
          Error(PNOCQUOTE, "");
          break;
        }
        while (c[i] != '\"')
        {
          t++;
          if (c[i] == '\\')
          {
            switch (c[++i])
            {
              case '\\':
                n = '\\';
                break;
              case '\'':
                n = '\'';
                break;
              case '"':
                n = '"';
                break;
              case '0':
                n = 0;
                break;
              default:
                sym[0] = c[i];
                sym[1] = '\0';
                Error(PINVESC, sym);
            }
          }
          else
            n = c[i];
          if (output)
            objectfile.OutputByte((char)n);
          i++;
        }
        if ((e = strchr(&c[i], ',')) != NULL)
          i = (e - c) + 1;
        else
          break;
      }
      else
      {
        t++;
        if ((e = strchr(&c[i], ',')) != NULL)
          *e = '\0';
        n = Parse(&c[i], NULL, 0);
        if (External(&c[i], false) == true)
          Error(PEXTDD, "");
        if (n > 255)
          Warning(WPTRUNCB, "");
        if (output)
          objectfile.OutputByte((char)(n % 256));
        if (e != NULL)
          i = (e - c) + 1;
        else
          break;
      }
    }
  }
  else if ((!stricmp(c, "dw")) || (!stricmp(c, "word")))
  {
    while (c[i] != '\0')
    {
      if ((e = strchr(&c[i], ',')) != NULL)
        *e = '\0';
      if (!addsym)
      {
        n = Parse(&c[i], NULL, 0);
        if (Local(&c[i], false))
          objectfile.AddLocalAddr(assembler.PCounter() + t);
        if (External(&c[i], false) == true)
          Error(PEXTDD, "");
        if (n > 65535)
          Warning(WPTRUNCW, "");
      }
      t += 2;
      if (output)
      {
        objectfile.OutputByte((char)(n % 256));
        objectfile.OutputByte((char)(n >> 8));
      }
      if (e != NULL)
        i = (e - c) + 1;
      else
        break;
    }
  }
  else if (!stricmp(c, "pad"))
  {
    n = Parse(&c[i], NULL, 0);
    if ((Local(&c[i], false)) || (External(&c[i], false)) ||
        (ExtRef(&c[i], false)))
    {
      Error(PPADNOTCONST, "");
      return 0;
    }
    if (n < _pcounter)
    {
      Error(PNOPADBACK, "");
      return 0;
    }
    t = n - _pcounter;
    if (output)
      for(i = _pcounter; i < n; i++)
        objectfile.OutputByte(0xff);
  }
  else if (!stricmp(c, "incbin"))
  {
    if (c[i] == '\"')
    {
      i++;
      if ((e = strchr(&c[i], '\"')) == NULL)
        Error(PNOCQUOTE, "");
      else
        *e = '\0';       
    }
    fp = fopen(&c[i], "rb");
    if (fp == NULL)
    {
      Error(PNOOPENBIN, &c[i]);
      return 0;
    }
    fseek(fp, 0L, SEEK_END);
    n = ftell(fp);
    if (output)
    {
      fseek(fp, 0L, SEEK_SET);
      e = new char[n];
      fread(e, n, 1, fp);
      for(i = 0; i < n; i++)
        objectfile.OutputByte(e[i]);
      delete []e;
    }
    fclose(fp);
    t = n;
  }
  else if ((!stricmp(c, "dbs")) || (!stricmp(c, "bytes")))
  {
    n = Parse(&c[i], NULL, 0);
    if ((Local(&c[i], false)) || (External(&c[i], false)) ||
        (ExtRef(&c[i], false)))
    {
      Error(PDDNOTCONST, "");
      return 0;
    }
    t = n;
    if (output)
      for(i = 0; i < n; i++)
        objectfile.OutputByte(0xff);
  }
  else if ((!stricmp(c, "dws")) || (!stricmp(c, "words")))
  {
    n = Parse(&c[i], NULL, 0);
    if ((Local(&c[i], false)) || (External(&c[i], false)) ||
        (ExtRef(&c[i], false)))
    {
      Error(PDDNOTCONST, "");
      return 0;
    }
    n *= 2;
    t = n;
    if (output)
      for(i = 0; i < n; i++)
        objectfile.OutputByte(0xff);
  }
  else if (!stricmp(c, "end"))
    return -1;
  else
    Error(PUNREQDIR, c);

  return t;
}

